/*-
 * Copyright (C) 2008 A.Leo.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * The driver for the m41t80 RTC. It could be found in a DNS-323 NAS
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/conf.h>

#include <dev/clock_subr.h>

#include <dev/i2c/i2cvar.h>
#include <dev/i2c/m41t80reg.h>

struct m41t80rtc_softc {
	struct device sc_dev;
	i2c_tag_t sc_tag;
	int sc_address;
	int sc_open;
	struct todr_chip_handle sc_todr;
};

static void	m41t80rtc_attach(struct device *, struct device *, void *);
static int	m41t80rtc_match(struct device *, struct cfdata *, void *);

CFATTACH_DECL(m41t80rtc, sizeof(struct m41t80rtc_softc),
    m41t80rtc_match, m41t80rtc_attach, NULL, NULL);

static int m41t80rtc_clock_read(struct m41t80rtc_softc *, struct clock_ymdhms *);
static int m41t80rtc_clock_write(struct m41t80rtc_softc *, struct clock_ymdhms *);
static int m41t80rtc_gettime(struct todr_chip_handle *, volatile struct timeval *);
static int m41t80rtc_settime(struct todr_chip_handle *, volatile struct timeval *);

static int
m41t80rtc_match(struct device *parent, struct cfdata *cf, void *arg)
{
	struct i2c_attach_args *ia = arg;

	if (ia->ia_addr == M41T80_ADDR)
		return 1;

	return 0;
}

static void
m41t80rtc_attach(struct device *parent, struct device *self, void *arg)
{
	struct m41t80rtc_softc *sc = device_private(self);
	struct i2c_attach_args *ia = arg;

	aprint_naive(": Real-time Clock\n");
	aprint_normal(": M41T80 Real-time Clock\n");

	sc->sc_tag = ia->ia_tag;
	sc->sc_address = ia->ia_addr;
	sc->sc_open = 0;
	sc->sc_todr.cookie = sc;
	sc->sc_todr.todr_gettime = m41t80rtc_gettime;
	sc->sc_todr.todr_settime = m41t80rtc_settime;
	sc->sc_todr.todr_setwen = NULL;

	todr_attach(&sc->sc_todr);
}


static int
m41t80rtc_gettime(struct todr_chip_handle *ch, volatile struct timeval *tv)
{
	struct m41t80rtc_softc *sc;
	struct clock_ymdhms dt;

	sc = ch->cookie;
	memset(&dt, 0, sizeof(dt));

	m41t80rtc_clock_read(sc, &dt);

	tv->tv_sec = clock_ymdhms_to_secs(&dt);
	tv->tv_usec = 0;

	return 0;
}

static int
m41t80rtc_settime(struct todr_chip_handle *ch, volatile struct timeval *tv)
{
	struct m41t80rtc_softc *sc = ch->cookie;
	struct clock_ymdhms dt;

	clock_secs_to_ymdhms(tv->tv_sec, &dt);

	if (m41t80rtc_clock_write(sc, &dt) == 0)
		return -1;

	return 0;
}

static int
m41t80rtc_clock_read(struct m41t80rtc_softc *sc, struct clock_ymdhms *dt)
{
	uint8_t bcd[M41T80_DATE_BYTES], cmdbuf[1];
	int error_code;

	if (iic_acquire_bus(sc->sc_tag, I2C_F_POLL)) {
		printf("%s: m41t80rtc_clock_read: failed to acquire I2C bus\n",
		    sc->sc_dev.dv_xname);
		return (0);
	}

	cmdbuf[0] = 0;
	error_code = iic_exec(sc->sc_tag, I2C_OP_READ_WITH_STOP,
			sc->sc_address,
			cmdbuf, 1,
			bcd, M41T80_DATE_BYTES,
			I2C_F_POLL);

	iic_release_bus(sc->sc_tag, I2C_F_POLL);


	if (error_code) {
		printf("%s: failed to read data, error code %i\n",
				sc->sc_dev.dv_xname, error_code);
		return 0;
	}

	dt->dt_sec = FROMBCD(bcd[M41T80_REG_SEC] & M41T80_SEC_MASK);
	dt->dt_min = FROMBCD(bcd[M41T80_REG_MIN] & M41T80_MIN_MASK);
	dt->dt_hour = FROMBCD(bcd[M41T80_REG_CENHR] & M41T80_HOUR_MASK);
	dt->dt_day = FROMBCD(bcd[M41T80_REG_DATE] & M41T80_DATE_MASK);
	dt->dt_mon = FROMBCD(bcd[M41T80_REG_MONTH] & M41T80_MONTH_MASK);
	dt->dt_year = FROMBCD(bcd[M41T80_REG_YEAR]) + POSIX_BASE_YEAR;

	return 1;
}

static int
m41t80rtc_clock_write(struct m41t80rtc_softc *sc, struct clock_ymdhms *dt)
{
	uint8_t bcd[M41T80_DATE_BYTES], cmdbuf[2];
	int error_code;

	bcd[M41T80_REG_CSEC] = TOBCD(0);
	bcd[M41T80_REG_SEC] = TOBCD(dt->dt_sec);
	bcd[M41T80_REG_MIN] = TOBCD(dt->dt_min);
	bcd[M41T80_REG_CENHR] = TOBCD(dt->dt_hour);
	bcd[M41T80_REG_DATE] = TOBCD(dt->dt_day);
	bcd[M41T80_REG_DAY] = TOBCD(dt->dt_wday);
	bcd[M41T80_REG_MONTH] = TOBCD(dt->dt_mon);
	bcd[M41T80_REG_YEAR] = TOBCD((dt->dt_year - POSIX_BASE_YEAR) % 100);

	if (iic_acquire_bus(sc->sc_tag, I2C_F_POLL)) {
		printf("%s: m41t80rtc_clock_write: failed to acquire I2C bus\n",
		    sc->sc_dev.dv_xname);
		return 0;
	}

	cmdbuf[0] = M41T80_REG_SEC;
	cmdbuf[1] = M41T80_SEC_ST;

	if (iic_exec(sc->sc_tag, I2C_OP_WRITE, sc->sc_address,
		     cmdbuf, 1, &cmdbuf[1], 1, I2C_F_POLL)) {
		iic_release_bus(sc->sc_tag, I2C_F_POLL);
		printf("%s: m41t80rtc_clock_write: failed to Hold Clock\n",
		    sc->sc_dev.dv_xname);
		return 0;
	}


	cmdbuf[0] = 0;
	error_code = iic_exec(sc->sc_tag, I2C_OP_WRITE_WITH_STOP, sc->sc_address,
			cmdbuf, 1,
			bcd, M41T80_DATE_BYTES,
			I2C_F_POLL);

	iic_release_bus(sc->sc_tag, I2C_F_POLL);

	if (error_code) {
		printf("%s: failed to write clock, error code %i\n",
				sc->sc_dev.dv_xname, error_code);
		return 0;
	}

	return 1;
}
